/* 
   MPI Mandelbrot set computation

   Original program by Charles Calkins

   Modified by Jim Teresco, Williams College

   Uses MPI datatypes

   $Id$
*/

#include <string.h>
#include <stdlib.h>
#include <stdio.h>
#include <time.h>
#include <netdb.h>
#include <unistd.h>
#include <iostream>

#include <mpi.h>
#include "mand.h"

using namespace std;

// define a struct to contain all the info passed to a worker
typedef struct workunit {
  int miniy;
  int maxiy;
};

/* This version is better in the sense that it makes better choices
   for how to send and receive messages and it checks for errors.  But
   it's not as good in that so much stuff is right here in main() */
int main(int argc, char *argv[]) {
  int myid, nprocs;
  int rc;
  double xmin, xmax, ymin, ymax, h, thresh;
  int maxiter, nprocs_ignore, xsize, ysize;
  double *array;
  char filename[256];
  double delta;
  char hostname[MAXHOSTNAMELEN];

  /* initialize MPI */
  rc = MPI_Init(&argc, &argv);
  if (rc != MPI_SUCCESS) {
    cerr << "MPI_Init failed" << endl;
    MPI_Abort(MPI_COMM_WORLD, 1);
  }

  MPI_Comm_size(MPI_COMM_WORLD, &nprocs); 
  MPI_Comm_rank(MPI_COMM_WORLD, &myid);

  /* see where we're running */
  gethostname(hostname,MAXHOSTNAMELEN);
  
  /* say who we are */
  cout << "Rank " << myid << " of " << nprocs << " has pid " << getpid() << " on " << hostname << endl;

  /* the master process reads in any command-line arguments
     and prints them out */
  if (myid == 0) {
  
    parse_cmdline(argc, argv, 
		  &xsize, &ysize,
		  &xmin, &xmax, &ymin, &ymax, 
		  &maxiter, &h, &thresh, &nprocs_ignore, filename);

    delta=thresh*(xmax-xmin)/(xsize-1);
  
    printf("xsize=%d  ysize=%d\n", xsize, ysize);
    printf("xmin=%f  xmax=%f  ymin=%f  ymax=%f\n",
	   xmin, xmax, ymin, ymax);
    printf("maxiter=%d  h=%f  thresh=%f  nprocs=%d\n", 
	   maxiter, h, thresh, nprocs);
    printf("filename=%s\n", filename);
  }

  /* broadcast information that needs to be known everywhere */
  rc = MPI_Bcast(&xmin, 1, MPI_DOUBLE, 0, MPI_COMM_WORLD);
  if (rc != MPI_SUCCESS) {
    cerr << "MPI_Bcast failed" << endl;
    MPI_Abort(MPI_COMM_WORLD, 1);
  }
  rc = MPI_Bcast(&xmax, 1, MPI_DOUBLE, 0, MPI_COMM_WORLD);
  if (rc != MPI_SUCCESS) {
    cerr << "MPI_Bcast failed" << endl;
    MPI_Abort(MPI_COMM_WORLD, 1);
  }
  rc = MPI_Bcast(&ymin, 1, MPI_DOUBLE, 0, MPI_COMM_WORLD);
  if (rc != MPI_SUCCESS) {
    cerr << "MPI_Bcast failed" << endl;
    MPI_Abort(MPI_COMM_WORLD, 1);
  }
  rc = MPI_Bcast(&ymax, 1, MPI_DOUBLE, 0, MPI_COMM_WORLD);
  if (rc != MPI_SUCCESS) {
    cerr << "MPI_Bcast failed" << endl;
    MPI_Abort(MPI_COMM_WORLD, 1);
  }
  rc = MPI_Bcast(&xsize, 1, MPI_INT, 0, MPI_COMM_WORLD);
  if (rc != MPI_SUCCESS) {
    cerr << "MPI_Bcast failed" << endl;
    MPI_Abort(MPI_COMM_WORLD, 1);
  }
  rc = MPI_Bcast(&ysize, 1, MPI_INT, 0, MPI_COMM_WORLD);
  if (rc != MPI_SUCCESS) {
    cerr << "MPI_Bcast failed" << endl;
    MPI_Abort(MPI_COMM_WORLD, 1);
  }
  rc = MPI_Bcast(&maxiter, 1, MPI_INT, 0, MPI_COMM_WORLD);
  if (rc != MPI_SUCCESS) {
    cerr << "MPI_Bcast failed" << endl;
    MPI_Abort(MPI_COMM_WORLD, 1);
  }
  rc = MPI_Bcast(&h, 1, MPI_DOUBLE, 0, MPI_COMM_WORLD);
  if (rc != MPI_SUCCESS) {
    cerr << "MPI_Bcast failed" << endl;
    MPI_Abort(MPI_COMM_WORLD, 1);
  }
  rc = MPI_Bcast(&delta, 1, MPI_DOUBLE, 0, MPI_COMM_WORLD);
  if (rc != MPI_SUCCESS) {
    cerr << "MPI_Bcast failed" << endl;
    MPI_Abort(MPI_COMM_WORLD, 1);
  }

  double start=MPI_Wtime();
  
  /* define the MPI datatype for the "struct workunit"s we'll be
     passing */
  MPI_Datatype twoIntsType;
  rc = MPI_Type_contiguous(2, MPI_INT, &twoIntsType);
  if (rc != MPI_SUCCESS) {
    cerr << "MPI_Type_contiguous failed" << endl;
    MPI_Abort(MPI_COMM_WORLD, 1);
  }
  rc = MPI_Type_commit(&twoIntsType);
  if (rc != MPI_SUCCESS) {
    cerr << "MPI_Type_commit failed" << endl;
    MPI_Abort(MPI_COMM_WORLD, 1);
  }

  if (myid == 0) {
    /* the master doles out the work */
    array=new double[xsize*ysize];
    if (!array) {
      cerr << "Could not allocate memory for result" << endl;
      MPI_Abort(MPI_COMM_WORLD, 1);
    }

    int step;
    if (!nprocs)
      step=ysize;
    else
      step=ysize/(nprocs-1);
    
    int i;
    for (i=1; i<nprocs; i++) {
      struct workunit thework;
      thework.miniy = (i-1)*step;
      thework.maxiy = (i-1)*step+step-1;

      /* blocking send, perhaps should use nonblocking */
      rc = MPI_Send(&thework, 1, twoIntsType, i, 0, MPI_COMM_WORLD);
      if (rc != MPI_SUCCESS) {
	cerr << "MPI_Send failed" << endl;
	MPI_Abort(MPI_COMM_WORLD, 1);
      }
    }

    /* the workers all have their tasks...so we wait */

    /* the workers could finish in any order, but we know to expect
       exactly nprocs-1 sets of messages to come in */
    for (i=1; i<nprocs; i++) {
      struct workunit thework;
      MPI_Status status;

      rc = MPI_Recv(&thework, 1, twoIntsType, MPI_ANY_SOURCE, MPI_ANY_TAG,
		    MPI_COMM_WORLD, &status);
      if (rc != MPI_SUCCESS) {
	cerr << "MPI_Recv failed" << endl;
	MPI_Abort(MPI_COMM_WORLD, 1);
      }

      /* we don't know which process sent us this item, but we can check
	 in the status to find out -- that's the process whose actual
	 data we can expect now */
      int from = status.MPI_SOURCE;

      double *arraypart=new double[(thework.maxiy-thework.miniy+2)*xsize];
      rc = MPI_Recv(arraypart, (thework.maxiy-thework.miniy+2)*xsize, 
		    MPI_DOUBLE, from, MPI_ANY_TAG, MPI_COMM_WORLD, 
		    &status);
      if (rc != MPI_SUCCESS) {
	cerr << "MPI_Recv failed" << endl;
	MPI_Abort(MPI_COMM_WORLD, 1);
      }
      for (int iy=thework.miniy; iy<thework.maxiy; iy++)
	for (int ix=0; ix<xsize; ix++)
	  array[iy*xsize+ix]=arraypart[(iy-thework.miniy)*xsize+ix];
      delete arraypart;
    }
    /* all workers have reported in - the array is complete */
    cout << "Computation complete, elapsed time: " << 
      MPI_Wtime() - start << endl;

    /* this is where we would write out a graphics file or something */
  }
  else {
    /* we're a worker process - let's get our work, do it, report back */
    struct workunit thework;
    MPI_Status status;

    /* every worker gets one message with its range to work on */
    rc = MPI_Recv(&thework, 1, twoIntsType, 0, 0, MPI_COMM_WORLD,
		  &status);
    if (rc != MPI_SUCCESS) {
      cerr << "MPI_Recv failed" << endl;
      MPI_Abort(MPI_COMM_WORLD, 1);
    }

    /* do the work */
    double dist,cx,cy;
    int iy;
    
    /* yes, we're reusing the array pointer, but only the master has
       the real copy allocated, we are in a worker here */
    array=new double[(thework.maxiy-thework.miniy+2)*xsize];
    double *xorbit=new double[maxiter+1];
    double *yorbit=new double[maxiter+1];
    for (iy=thework.miniy; iy<=thework.maxiy; iy++) {
      cy=ymin+iy*(ymax-ymin)/(ysize-1);
      for (int ix=0; ix<xsize; ix++) {
	cx=xmin+ix*(xmax-xmin)/(xsize-1);
	dist=MSetDist(cx,cy,maxiter,h, xorbit, yorbit);
	if (dist<delta)
	  array[(iy-thework.miniy)*xsize+ix]=0;
	else
	  array[(iy-thework.miniy)*xsize+ix]=dist;		
      }
    }	
    
    delete xorbit;
    delete yorbit;

    /* send back the answers - first send the range back, so the
       master knows where to put the answers we have computed */

    rc = MPI_Send(&thework, 1, twoIntsType, 0, 0, MPI_COMM_WORLD);
    if (rc != MPI_SUCCESS) {
      cerr << "MPI_Send failed" << endl;
      MPI_Abort(MPI_COMM_WORLD, 1);
    }

    rc = MPI_Send(array, (thework.maxiy-thework.miniy+2)*xsize, 
		  MPI_DOUBLE, 0, 0, MPI_COMM_WORLD);
    if (rc != MPI_SUCCESS) {
      cerr << "MPI_Send failed" << endl;
      MPI_Abort(MPI_COMM_WORLD, 1);
    }

    delete array;
    /* worker done... */
  }

  MPI_Type_free(&twoIntsType);
  MPI_Finalize();
  return 0;
}

