/* 
   Mandelbrot set computation

   Original program by Charles Calkins

   Modified by Jim Teresco, Williams College

   $Id: mand.cc,v 1.1 2003/03/04 05:28:42 terescoj Exp $
*/

#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <math.h>
#include "mand.h"

#define ABS(x) ((x)<0?(-x):(x))

/* tsofi 198 */
double MSetDist(double cx, double cy, int maxiter, double hugeval, 
				double *xorbit, double *yorbit) {
  double x,y,x2,y2;
  double xder,yder, temp, dist;
  char flag;
  int iter,i;

  x=y=x2=y2=dist=xorbit[0]=yorbit[0]=0.0;
  iter=0;

  while ( (iter<maxiter) && (x2+y2<hugeval) ) {
    y=2*x*y+cy;
    x=x2-y2+cx;
    x2=x*x;
    y2=y*y;

    iter++;
    xorbit[iter]=x;
    yorbit[iter]=y;
  }

  if (x2+y2>hugeval) {
    xder=yder=0.0;
    i=0;
    flag=0;
    while ( (i<iter) && (!flag) ) {
      temp=2*(xorbit[i]*xder - yorbit[i]*yder)+1;
      yder=2*(yorbit[i]*xder+xorbit[i]*yder);
      xder=temp;
      if ((ABS(xder)>1.0E+60) || (ABS(yder)>1.0E+60))
        flag=1;
      i++;
    }
    if (!flag)
      dist=log(x2+y2)*sqrt(x2+y2)/sqrt(xder*xder+yder*yder);

  }

  
  return dist;
}


void parse_cmdline(int argc, char *argv[],
		   int *xsize, int *ysize,
		   double *xmin, double *xmax, double *ymin, double *ymax,
		   int *maxiter, double *h, double *thresh, int *nthreads,
		   char *filename) {
  int i=0;
  
  *xsize=256;
  *ysize=256;
  *xmin=-2.25;
  *xmax=0.75;
  *ymin=-1.5;
  *ymax=1.5;
  *maxiter=256;
  *h=10000000.0;
  *thresh=1.0;
  *nthreads=1;
  strcpy(filename, "mset.tga");
  
  if (argc>1) 
    while (i<argc) {
      if (!strcmp(argv[i], "-xsize")) {
	i++;
	if (i<argc) *xsize=atoi(argv[i]);
      }
      if (!strcmp(argv[i], "-ysize")) {
	i++;
	if (i<argc) *ysize=atoi(argv[i]);
      }
      if (!strcmp(argv[i], "-xmin")) {
	i++;
	if (i<argc) *xmin=atof(argv[i]);
      }
      if (!strcmp(argv[i], "-xmax")) {
	i++;
	if (i<argc) *xmax=atof(argv[i]);
      }
      if (!strcmp(argv[i], "-ymin")) {
	i++;
	if (i<argc) *ymin=atof(argv[i]);
      }
      if (!strcmp(argv[i], "-ymax")) {
	i++;
	if (i<argc) 
	  *ymax=atof(argv[i]);
      }
      if (!strcmp(argv[i], "-maxiter")) {
	i++;
	if (i<argc) 
	  *maxiter=atoi(argv[i]);
      }
      if (!strcmp(argv[i], "-h")) {
	i++;
	if (i<argc) 
	  *h=atof(argv[i]);
      }
      if (!strcmp(argv[i], "-thresh")) {
	i++;
	if (i<argc) 
	  *thresh=atof(argv[i]);
      }
      if (!strcmp(argv[i], "-nthreads")) {
	i++;
	if (i<argc) 
	  *nthreads=atoi(argv[i]);
      }
      if (!strcmp(argv[i], "-filename")) {
	i++;
	if (i<argc) 
	  strcpy(filename, argv[i]);
      }
      i++;
    }
  else {
    /* no arguments, assume want menu interface */
    i=0;
    
    while (i!=12) {
      printf("\n\n\n 1> xsize=%d\n", *xsize);
      printf(" 2> ysize=%d\n", *ysize);
      printf(" 3> xmin=%f\n", *xmin);
      printf(" 4> xmax=%f\n", *xmax);
      printf(" 5> ymin=%f\n", *ymin);
      printf(" 6> ymax=%f\n", *ymax);
      printf(" 7> maxiter=%d\n", *maxiter);
      printf(" 8> h=%f\n", *h);
      printf(" 9> threshold=%f\n", *thresh);
      printf("10> num threads=%d\n", *nthreads);
      printf("11> filename=%s\n", filename);
      printf("12> Generate\n");
      printf("\nChoice: ");
      scanf("%d", &i);
      
      switch (i) {
      case 1:
	printf("\nxsize=");
	scanf("%d", xsize);
	break;
      case 2:
	printf("\nysize=");
	scanf("%d", ysize);
	break;
      case 3:
	printf("\nxmin=");
	scanf("%lf", xmin);
	break;
      case 4:
	printf("\nxmax=");
	scanf("%lf", xmax);
	break;
      case 5:
	printf("\nymin=");
	scanf("%lf", ymin);
	break;
      case 6:
	printf("\nymax=");
	scanf("%lf", ymax);
	break;
      case 7:
	printf("\nmaxiter=");
	scanf("%d", maxiter);
	break;
      case 8:
	printf("\nh=");
	scanf("%lf", h);
	break;
      case 9:
	printf("\nthreshold=");
	scanf("%lf", thresh);
	break;
      case 10:
	printf("\nnum threads=");
	scanf("%d", nthreads);
	break;
      case 11:
	printf("\nfilename=");
	scanf("%s", filename);
	break;
      default:
	break;
      }
    }
    
  }
}




void SaveAsTGA(char *filename, int xsize, int ysize, double *array) {
  int x,y;
  TGAFH fh;
  FILE *f;
  BYTE r,g,b;
  double maximum;
  double val, v1, v2;
  
  maximum=-1E10;
  for (x=0; x<xsize; x++)
    for (y=0; y<ysize; y++)
      if (array[y*xsize+x]>maximum)
	maximum=array[y*xsize+x];
  
  memset(&fh, 0, sizeof(fh));
  
  fh.ImageType=2;
  fh.Width=(short)xsize;     
  fh.Height=(short)ysize;
  fh.PixelDepth=24;
  
  f=fopen(filename, "wb");
  /* if !f error */
  
  fwrite((char *)&fh.IDLength, sizeof(BYTE), 1, f); 
  fwrite((char *)&fh.ColorMapType, sizeof(BYTE), 1, f); 
  fwrite((char *)&fh.ImageType, sizeof(BYTE), 1, f); 
  fwrite((char *)&fh.CMapStart, sizeof(WORD), 1, f); 
  fwrite((char *)&fh.CMapLength, sizeof(WORD), 1, f); 
  fwrite((char *)&fh.CMapDepth, sizeof(BYTE), 1, f); 
  fwrite((char *)&fh.XOffset, sizeof(WORD), 1, f); 
  fwrite((char *)&fh.YOffset, sizeof(WORD), 1, f); 
  fwrite((char *)&fh.Width, sizeof(WORD), 1, f); 
  fwrite((char *)&fh.Height, sizeof(WORD), 1, f); 
  fwrite((char *)&fh.PixelDepth, sizeof(BYTE), 1, f); 
  fwrite((char *)&fh.ImageDescriptor, sizeof(BYTE), 1, f); 
  
  
  for (y=0; y<ysize; y++)
    for (x=0; x<xsize; x++) {
      int offset = y*xsize+x;
      
      
      if (array[offset]<1E-4)
	r=g=b=255;
      else {
	r=g=b=0;
	val=array[offset]/maximum;
	v1=sqrt(sqrt(val))/2.0;
	r=255-((BYTE)(v1*255));
	v2=sqrt(sqrt(val))/1;
	g=255-((BYTE)(v2*255));
      }
      
      fwrite((char *)&b, 1, 1, f);  // blue
      fwrite((char *)&g, 1, 1, f);  // green
      fwrite((char *)&r, 1, 1, f);  // red
      
    }
  
  fclose(f);
}


